#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "mixasm.h"

Symbol  *symlist = 0;    /* symbol table: linked list */

Symbol *
lookup(char *s)        /* find s in symbol table */
{
    Symbol  *sp;

    for (sp = symlist; sp != (Symbol *) 0; sp = sp->next)
	if (strcmp(sp->name, s) == 0)
	    return sp;
    return 0;        /* 0 ==> not found */
}

Symbol *
install(char *s, int v, int l) /* install s in symbol table */
{
    Symbol *sp;
    char *emalloc();

    sp = (Symbol *) emalloc(sizeof(Symbol));
    sp->name = emalloc(strlen(s)+1);  /* +1 for '\0' */
    strcpy(sp->name, s);
    sp->value = v;
    sp->line = l;    
    sp->reflist = 0;
    sp->next = symlist;  /* put at front of list */
    symlist = sp;
    return sp;
}

char *
emalloc(unsigned n)          /* check return from malloc */
{
    char *p;

    p = malloc(n);
    if (p == 0)
	 yyerror("out of memory");
    return p;
}

        
ref *
addref(ref *p, int loc, int line)
{
    ref *sp;

    sp = (ref *) emalloc(sizeof(ref));
    sp->next = p;
    sp->address = loc;
    sp->line = line;
    return sp;
}

/* The following is based on a quicksort routine for linked-lists
 * by Jon Guthrie.
 */
#define getnext(a) ((a)->next)
#define setnext(a,l) (a->next = l, l = a)

List *
sortl(List *list, int (*compare)())
{
    List *low_list, *high_list, *current, *pivot, *temp;
    int     result;

    if (!list)
        return NULL;

    /*
     * Find the first element that doesn't have the same value as the first
     * element.
     */
    current = list;
    do {
        current = getnext(current);
        if (!current)
            return list;
    } while (0 == (result = compare(list, current)));

    /*
     *  My pivot value is the lower of the two.  This insures that the sort
     *  will always terminate by guaranteeing that there will be at least one
     *   member of both of the sublists.
     */
    if (result > 0)
        pivot = current;
    else
        pivot = list;

    /* Initialize the sublist pointers */
    low_list = high_list = NULL;

    /*
     *  Now, separate the items into the two sublists
     */
    current = list;
    while (current) {
        temp = getnext(current);
        if (compare(pivot, current) < 0) {
            /* add one to the high list */
            setnext(current, high_list);
            high_list = current;
        } else {
            /* add one to the low list */
            setnext(current, low_list);
            low_list = current;
        }
        current = temp;
    }
    
    /* And, recursively call the sort for each of the two sublists. */
    low_list  = sortl(low_list, compare);
    high_list = sortl(high_list, compare);

    /*
     * Now, I have to put the "high" list after the end of the "low" list.  
     * To do that, I first have to find the end of the "low" list...
     */
    current = temp = low_list;
    while (1) {
        current = getnext(current);
        if (!current)
            break;
        temp = current;
    }

    /* Then, I put the "high" list at the end of the low list */
    setnext(temp, high_list);
    return low_list;
}

int
compsym(Symbol *a, Symbol *b)
{
    return a->value - b->value;
}

int
compref(struct ref *a, struct ref *b)
{
    return a->line - b->line;
}

void
sort_symbols()
{
    symlist = (Symbol*)sortl((List*)symlist, compsym);
}

void
sort_refs(Symbol *s)
{
    s->xref = (struct ref *)sortl((List*)s->xref, compref);
}
